/**
 * Copyright (c) 2019 Eclipse contributors and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v2.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v20.html
 */
package org.eclipse.oomph.download.index;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.Reader;
import java.math.RoundingMode;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.stream.Stream;

public class DownloadIndex {

	private final String site;

	private final String targetOutput;

	private final boolean trace = "true"
			.equals(System.getProperty("org.eclipse.oomph.download.index.DownloadIndex.trace"));

	private final boolean traceFile = "true"
			.equals(System.getProperty("org.eclipse.oomph.download.index.DownloadIndex.traceFile"));

	private final String currentDate;

	private final String currentTime;

	private final Map<String, String> folderData = Collections.synchronizedMap(new LinkedHashMap<String, String>());

	private final Map<String, String> folderSizes = new ConcurrentHashMap<String, String>();

	private final Map<String, Integer> fileCounts = new ConcurrentHashMap<String, Integer>();

	private final Map<String, String> fileSizes = new ConcurrentHashMap<String, String>();

	private ExecutorService executorService;

	public DownloadIndex(String site, String targetOutput) {
		this.site = site;
		this.targetOutput = targetOutput;
		long now = System.currentTimeMillis();
		currentDate = new SimpleDateFormat("YYYY-MM-dd").format(now);
		currentTime = new SimpleDateFormat("HH:MM").format(now);
	}

	public static void main(String[] args) {
		try {
			new DownloadIndex(args[0], args[1]).run();
		} catch (Exception ex) {
			ex.printStackTrace();
			System.exit(-1);
		}
	}

	private void run() throws Exception {
		executorService = Executors.newFixedThreadPool(30);

		handle("", site).call();
		executorService.shutdown();
		executorService.awaitTermination(1, TimeUnit.MINUTES);

		PrintStream out = new PrintStream(targetOutput, "UTF-8");
		int length = Paths.get(site).toUri().toString().length();
		for (Entry<String, String> entry : folderData.entrySet()) {
			String location = entry.getKey();
			String size = folderSizes.get(location);
			int fileCount = fileCounts.get(location);
			String fileSize = fileSizes.get(location);
			String relativeLocation = Paths.get(location).toUri().toString().substring(length).replaceAll("/$", "");

			if (relativeLocation.isEmpty()) {
				relativeLocation = "/";
			}

			out.println(size + "\t" + relativeLocation + "\t" + entry.getValue() + "\t" + fileCount + "\t" + fileSize);

			System.out.println(
					size + "\t" + relativeLocation + "\t" + entry.getValue() + "\t" + fileCount + "\t" + fileSize);
		}

		out.close();
	}

	public static float parseSize(String size) {
		float result;
		size = size.replace(",", "");
		if (size.endsWith("K")) {
			result = Float.parseFloat(size.substring(0, size.length() - 1)) * 1024f;
		} else if (size.endsWith("M")) {
			result = Float.parseFloat(size.substring(0, size.length() - 1)) * 1024f * 1024f;
		} else if (size.endsWith("G")) {
			result = Float.parseFloat(size.substring(0, size.length() - 1)) * 1024f * 1024f * 1024f;
		} else if (size.endsWith("T")) {
			result = Float.parseFloat(size.substring(0, size.length() - 1)) * 1024f * 1024f * 1024f * 1024f;
		} else {
			result = Float.parseFloat(size);
		}
		return result;
	}

	public static String format(float size) {
		NumberFormat instance = NumberFormat.getInstance(Locale.US);
		instance.setMaximumFractionDigits(1);
		instance.setRoundingMode(RoundingMode.DOWN);
		if (size < 1024f) {
			return instance.format(size);
		} else if (size < 1024f * 1024f) {
			return instance.format(size / 1024) + "K";
		} else if (size < 1024f * 1024f * 1024f) {
			return instance.format(size / 1024f / 1024f) + "M";
		} else if (size < 1024f * 1024f * 1024f * 1024f) {
			return instance.format(size / 1024f / 1024f / 1024f) + "G";
		} else {
			return instance.format(size / 1024f / 1024f / 1024f / 1024f) + "T";
		}
	}

	private Callable<Float> handle(String indent, String location)
			throws IOException, InterruptedException, ExecutionException {

		if (indent.isEmpty()) {
			folderData.put(location, currentDate + " " + currentTime);
		}

		int fileCount = 0;
		float totalFileSize = 0.0f;

		List<Future<Callable<Float>>> folderFutures = new ArrayList<Future<Callable<Float>>>();

		Path locationPath = Paths.get(location);
		Stream<Path> stream = Files.list(locationPath);
		Path[] list = stream.toArray(Path[]::new);
		stream.close();
		for (Path path : list) {
			if (!Files.isSymbolicLink(path)) {
				long lastModifiedTime = Files.getLastModifiedTime(path).toMillis();
				String date = new SimpleDateFormat("YYYY-MM-dd").format(lastModifiedTime);
				String time = new SimpleDateFormat("HH:MM").format(lastModifiedTime);
				if (Files.isRegularFile(path)) {
					float size = Files.size(path);
					float byteSize = size;

					if (traceFile) {
						System.out.println(indent + location + "' date='" + date + "' time='" + time + "' size='" + size
								+ "' ? " + format(byteSize));
					}

					totalFileSize += byteSize;
					++fileCount;
				} else if (Files.isDirectory(path) && Files.isReadable(path)) {
					if (trace) {
						System.out.println(indent + "> " + path + "' date='" + date + "' time='" + time);
					}

					String pathString = path.toString().replace('\\', '/');
					folderData.put(pathString, date + " " + time);

					Future<Callable<Float>> folderFuture = executorService
							.submit(() -> handle(indent + "  ", pathString));
					folderFutures.add(folderFuture);

					if (trace) {
						System.out.println(indent + "< " + path + "' date='" + date + "' time='" + time);
					}
				}
			}
		}

		fileCounts.put(location, fileCount);
		fileSizes.put(location, format(totalFileSize));

		float totalSoFar = totalFileSize;
		return () -> {
			float finalTotal = totalSoFar;
			for (Future<Callable<Float>> folderFuture : folderFutures) {
				finalTotal += folderFuture.get().call();
			}

			folderSizes.put(location, format(finalTotal));
			return finalTotal;
		};
	}

	public static List<String> readLines(String location) throws IOException, InterruptedException {
		for (int i = 0; true; ++i) {
			try {
				URL url = new URL(location);
				InputStream inputStream = url.openStream();
				List<String> lines = readLines(inputStream, "UTF-8");
				inputStream.close();
				return lines;
			} catch (Exception ex) {
				if (i < 3) {
					System.err.println(ex.getLocalizedMessage());
					Thread.sleep(1000 * 60);
				} else {
					ex.printStackTrace();
					return Collections.emptyList();
				}
			}
		}
	}

	public static List<String> readLines(File location) throws IOException {
		InputStream in = new FileInputStream(location);
		List<String> lines = DownloadIndex.readLines(in, "UTF-8");
		in.close();
		return lines;

	}

	public static List<String> readLines(InputStream in, String charsetName) throws IOException {
		List<String> lines = new ArrayList<String>();

		Reader reader = null;
		BufferedReader bufferedReader = null;

		try {
			reader = new InputStreamReader(in, charsetName);
			bufferedReader = new BufferedReader(reader);

			String line;
			while ((line = bufferedReader.readLine()) != null) {
				lines.add(line);
			}
		} finally {
			if (bufferedReader != null) {
				bufferedReader.close();
			}
			if (reader != null) {
				reader.close();
			}
		}

		return lines;
	}
}
