/**
 * Copyright (c) 2019 Eclipse contributors and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v2.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v20.html
 */
package org.eclipse.oomph.download.index;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class DownloadIndexGenerator {

	/**
	 * This is where the various Oomph reports are kept that have huge file trees,
	 * including this index report's tree.
	 */
	private Pattern SELF_EXCLUSION_PATTERN = Pattern.compile("^oomph/archive/[^/]+/");

	private static String FILES = "%FILES%";

	private File outputFolder;

	private String summaryLocation;

	private Map<String, String> sizes = new LinkedHashMap<String, String>();

	private Map<String, String> dates = new LinkedHashMap<String, String>();

	private Map<String, String> times = new LinkedHashMap<String, String>();

	private Map<String, String> fileCounts = new LinkedHashMap<String, String>();

	private Map<String, List<String>> children = new LinkedHashMap<String, List<String>>();

	private float totalSize;

	private String linkURL;

	private String rootName;

	public DownloadIndexGenerator(String summaryLocation, String linkURL, String rootName) {
		this.summaryLocation = summaryLocation;
		this.linkURL = linkURL;
		this.rootName = rootName.intern();
		outputFolder = new File(summaryLocation).getAbsoluteFile().getParentFile();

	}

	public static void main(String[] args) throws IOException, InterruptedException {
		new DownloadIndexGenerator(args[0], args.length < 2 ? "https://download.eclipse.org/justj/?file=" : args[1],
				args.length < 3 ? "eclipse" : args[2]).run();
	}

	private void run() throws IOException, InterruptedException {
		List<String> lines = DownloadIndex.readLines(new File(summaryLocation));
		for (String line : lines) {
			String[] parts = line.split("\\s");
			String size = parts[0];
			String baseLocation = parts[1];
			if (!SELF_EXCLUSION_PATTERN.matcher(baseLocation).find()) {
				String location = baseLocation.length() == 1 ? rootName : rootName + "/" + baseLocation;
				location = location.replace("/./", "");
				String date = parts[2];
				String time = parts[3];
				String fileCount = parts[4];
				String fileSize = parts[5];
				handle(location, size, date, time, fileCount, fileSize);
			}
		}

		Comparator<String> sizeComparator = new Comparator<String>() {
			@Override
			public int compare(String o1, String o2) {
				String size1 = sizes.get(o1);
				String size2 = sizes.get(o2);
				return Float.compare(DownloadIndex.parseSize(size2), DownloadIndex.parseSize(size1));
			}
		};

		for (List<String> list : children.values()) {
			Collections.sort(list, sizeComparator);
		}

		List<String> children = this.children.get(rootName);

		totalSize = DownloadIndex.parseSize(sizes.get(rootName));

		visit(rootName, children, null, null);
	}

	private static String formatPercent(float numerator, float denominator) {
		if (denominator == 0f) {
			return "100.00%";
		} else {
			float value = 100f * numerator / denominator;
			NumberFormat instance = NumberFormat.getInstance(Locale.US);
			instance.setMaximumFractionDigits(2);
			instance.setMinimumFractionDigits(2);
			return instance.format(value) + '%';
		}
	}

	private void visit(String location, List<String> children, PrintStream out, String parentDirectLinkPrefix)
			throws IOException {
		String[] segments = location.split("/", 0);
		String name = segments[segments.length - 1];
		boolean isFiles = FILES.equals(name);
		String size = sizes.get(location);

		// Print this folder's index page.
		String cleansedName = htmlEscape(cleanse(name));
		if (!isFiles) {
			File file = new File(outputFolder, uriDecode(location));
			if (!file.mkdir() && !file.isDirectory()) {
				System.err.println("###" + file);
			}

			File index = new File(file, "index.html");
			PrintStream indexOut = new PrintStream(index, "UTF-8");
			String directLinkPrefix = printHTMLPrefix(indexOut, location);

			if (out != null) {
				// Print the parent entry.
				indexOut.println(//
						"  <tr><td><a href='../index.html#" + cleansedName + "'>..</a></td>" + //
								"<td>" + size + "</td>" + //
								"<td></td>" + //
								"<td>" + dates.get(location) + "</td></tr>");
			}

			for (String child : children) {
				List<String> grandChildren = this.children.get(child);
				visit(child, grandChildren, indexOut, directLinkPrefix);
			}

			printHTMLSuffix(indexOut);

			indexOut.close();

			if (Boolean.FALSE) {
				System.err.println("###" + index);
				String content = Files.readString(index.toPath(), StandardCharsets.UTF_8);

				Pattern pattern = Pattern.compile("<td><a[^>]+>([^<]+)</a></td><td[^>]*>([^<]+)</td>");
				for (Matcher matcher = pattern.matcher(content); matcher.find();) {
					System.err.println("dir-name=" + matcher.group(1) + " size=" + matcher.group(2));
				}
			}
		}

		// Print into the parent folder's table of contents.
		if (out != null) {
			String parentLocation = location.substring(0, location.lastIndexOf('/'));
			String parentSize = sizes.get(parentLocation);
			float parentSizeValue = DownloadIndex.parseSize(parentSize);
			float sizeValue = DownloadIndex.parseSize(size);
			String formattedPercent = formatPercent(sizeValue, parentSizeValue);
			String specializedSizeStyle = size.endsWith("G") || size.endsWith("T") ? " class='big'" : "";

			String label = isFiles
					? parentDirectLinkPrefix + "<em>Local Files: " + fileCounts.get(location) + "</em></a>"
					: "<a id='" + cleansedName + "' href='" + htmlEscape(name) + "/index.html'>"
							+ htmlEscape(uriDecode(name)) + "</a>";

			out.println(//
					"  <tr><td>" + label + "</td>" + //
							"<td" + specializedSizeStyle + ">" + size + "</td>" + //
							"<td>" + formattedPercent + "</td>" + //
							"<td>" + dates.get(location) + "</td></tr>");
		}
	}

	private String cleanse(String location) {
		return location.replace(':', '_');
	}

	public String uriDecode(String value) {
		return URLDecoder.decode(value.replace("+", "%2B"), StandardCharsets.UTF_8);
	}

	public String htmlEscape(String value) {
		if (value == null) {
			return "null";
		}
		String result = value.replace("&", "&amp;").replace("<", "&lt;").replace("\"", "&quote;");
		return result;
	}

	private String printHTMLPrefix(PrintStream out, String location) {
		class BreadCrumb {
			public String name;
			public String size;
			public String percent;
			public String eclipseRelativePath;
			public String navigationHref = "";
			public String rootHref = "..";

			public BreadCrumb(String name, String size) {
				this.name = name;
				this.size = size;
				if (name == rootName) {
					this.percent = "";
					this.eclipseRelativePath = "";
				} else {
					this.percent = formatPercent(DownloadIndex.parseSize(size), totalSize);
					this.eclipseRelativePath = name;
				}
			}

			public void addTo(List<BreadCrumb> breadCrumbs) {
				for (int i = 0, size = breadCrumbs.size(); i < size; i++) {
					BreadCrumb breadCrumb = breadCrumbs.get(i);
					if (i == 0) {
						navigationHref = htmlEscape(breadCrumb.rootHref + "/index.html#" + cleanse(breadCrumb.name));
						rootHref = breadCrumb.rootHref = "../" + breadCrumb.rootHref;
					} else {
						breadCrumb.rootHref = "../" + breadCrumb.rootHref;
					}

					if (name != rootName) {
						breadCrumb.eclipseRelativePath = name + "/" + breadCrumb.eclipseRelativePath;
					}
				}

				breadCrumbs.add(0, this);
			}

			@Override
			public String toString() {
				return "BreadCrumb [name=" + name + ", size=" + size + ", percent=" + percent + ", eclipseRelativePath="
						+ eclipseRelativePath + ", navigationHref=" + navigationHref + ", rootHref=" + rootHref + "]";
			}

			public String getMirorrLinkPrefix() {
				return "<a href='" + htmlEscape(linkURL + eclipseRelativePath) + "' target='eclipse_direct'>";
			}

			public String toHTML(boolean isDirectLink) {
				return getMirorrLinkPrefix() + "<img src='" + rootHref + "/Nav.png'/></a> "
						+ (isDirectLink ? getMirorrLinkPrefix() + uriDecode(name) + "</a>"
								: "<a href='" + navigationHref + "'>" + uriDecode(name) + "</a>") //
						+ " <small class='cs'>" + size + "</small>" //
						+ " <small>"
						+ (eclipseRelativePath.isEmpty() ? "" : (percent.equals("0.00%") ? "&lt; 0.01%" : "" + percent))
						+ "</small>";
			}
		}

		String baseLocation = location;
		List<BreadCrumb> breadCrumbs = new ArrayList<BreadCrumb>();
		for (int index = baseLocation.lastIndexOf('/'); index != -1; index = baseLocation.lastIndexOf('/')) {
			String baseSize = sizes.get(baseLocation);
			String childName = baseLocation.substring(index + 1);

			new BreadCrumb(childName, baseSize).addTo(breadCrumbs);

			baseLocation = baseLocation.substring(0, index);
		}

		new BreadCrumb(rootName, sizes.get(rootName)).addTo(breadCrumbs);
		out.println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">");
		out.println("<html>");
		out.println("<head>");
		out.println("  <title>");
		out.println("    " + location);
		out.println("  </title>");
		out.println("  <link rel='stylesheet' type='text/css' href='" + breadCrumbs.get(breadCrumbs.size() - 1).rootHref
				+ "/index.css'>");
		out.println("</head>");
		out.println("<body>");
		out.println("<h3>");

		for (int i = 0, limit = breadCrumbs.size() - 1; i <= limit; i++) {
			BreadCrumb breadCrumb = breadCrumbs.get(i);
			String html = "<span>" + breadCrumb.toHTML(i == limit) //
					+ (i == limit
							? " <small class='small'>&bull; " + dates.get(rootName) + " " + times.get(rootName)
									+ "</small>"
							: "") //
					+ "</span>";
			out.println("    " + html);
		}

		out.println("</h3>");
		out.println("<table>");
		out.println("  <tr><th>Name</th>" + "<th>Size</th>" + "<th>Percent</th>" + "<th>Date</th></tr>");

		return breadCrumbs.get(breadCrumbs.size() - 1).getMirorrLinkPrefix();
	}

	private void printHTMLSuffix(PrintStream out) {
		out.println("</table>");
		out.println("</body>");
	}

	private void handle(String location, String size, String date, String time, String fileCount, String fileSize) {
		sizes.put(location, size);
		dates.put(location, date);
		times.put(location, time);

		int index = location.lastIndexOf('/');
		if (index != -1) {
			String parent = location.substring(0, index);
			children.get(parent).add(location);
		}

		List<String> children = new ArrayList<String>();
		if (!"0".equals(fileCount)) {
			String files = location + "/" + FILES;
			sizes.put(files, fileSize);
			dates.put(files, "");
			times.put(files, "");
			fileCounts.put(files, fileCount);
			children.add(files);
		}

		this.children.put(location, children);
	}
}
